import boto3

from prismarine.runtime.dynamo_access import DynamoAccess


class DefaultDynamoAccess(DynamoAccess):
    def get_resource(self):
        return boto3.resource('dynamodb')

    def get_table(self, full_model_name: str):
        return self.get_resource().Table(full_model_name)


dynamoaccess = DefaultDynamoAccess()


def get_dynamo_access():
    return dynamoaccess



from decimal import Decimal
from typing import Any, Literal, TypeVar
import uuid
from boto3.dynamodb.conditions import Key

from prismarine.runtime.dynamo_access import DynamoAccess


# Main point for now is to remove any Decimals from the data
# since they are not JSON serializable and dont work with timedelta
def prepare_item(val: Any):
    if isinstance(val, list):
        return [prepare_item(v) for v in val]
    elif isinstance(val, dict):
        return {k: prepare_item(v) for k, v in val.items()}
    elif isinstance(val, Decimal):
        if val % 1 == 0:
            return int(val)
        else:
            return float(val)

    return val


def serialize_item(val: Any):
    if isinstance(val, list):
        return [serialize_item(v) for v in val]
    elif isinstance(val, dict):
        return {k: serialize_item(v) for k, v in val.items()}
    elif isinstance(val, int) and val is not True and val is not False:
        return Decimal(val)
    elif isinstance(val, float):
        return Decimal(str(val))

    return val


def _query(
    dynamo: DynamoAccess,
    table: str,
    kv: dict[str, str | int],
    index: str = '',
    limit: int | None = None,
    direction: Literal['ASC'] | Literal['DESC'] = 'ASC',
    **kwargs
) -> list[Any]:
    if len(kv) < 1 or len(kv) > 2:
        raise Exception(f'Invalid query params {kv}')

    keys = list(kv.keys())

    condition = Key(keys[0]).eq(kv[keys[0]])
    if len(kv) == 2:
        condition = condition & Key(keys[1]).eq(kv[keys[1]])

    query_args: dict[str, Any] = {'KeyConditionExpression': condition}
    if index:
        query_args['IndexName'] = index

    if limit:
        query_args['Limit'] = limit

    if direction == 'DESC':
        query_args['ScanIndexForward'] = False

    query_args.update(kwargs)

    return prepare_item(dynamo.get_table(table).query(**query_args).get('Items', []))  # type: ignore


def _get_item(
    dynamo: DynamoAccess,
    table: str,
    kv: dict[str, str | int],
    default: Any = ...,
    **kwargs
) -> Any | None:
    return prepare_item(
        dynamo.get_table(table)
        .get_item(Key=kv, **kwargs)
        .get('Item', default if default is not ... else None)
    )


def _put_item(
    dynamo: DynamoAccess,
    table: str,
    item: Any,
    **kwargs
):
    dynamo.get_table(table).put_item(Item=serialize_item(item), **kwargs)  # type: ignore


SaveItem = TypeVar('SaveItem', bound=Any)


def _save(
    dynamo: DynamoAccess,
    table: str,
    *,
    kv: dict[str, str | int],
    updated: SaveItem,
    original: SaveItem | None = None,
    **kwargs
):
    if not original:
        original = _get_item(dynamo, table, kv, **kwargs)

    if not original:
        _put_item(dynamo, table, updated)
        return

    diff = {
        key: updated[key] for key in updated.keys() if updated[key] != original[key]
    }
    if len(diff) == 0:
        return

    _update(dynamo, table, kv, diff)


def _update(
    dynamo: DynamoAccess,
    table: str,
    kv: dict[str, str | int],
    item: Any,
    **kwargs
):
    vals = {}
    keys = {}
    expressions = []
    for k, v in item.items():
        vals[f':{k}'] = serialize_item(v)
        keys[f'#{k}'] = k
        expressions.append(f'#{k} = :{k}')

    update_expression = 'SET ' + ', '.join(expressions)

    dynamo.get_table(table).update_item(
        Key=kv,
        UpdateExpression=update_expression,
        ExpressionAttributeNames=keys,
        ExpressionAttributeValues=vals,
        **kwargs
    )


def _delete(
    dynamo: DynamoAccess,
    table: str,
    kv: dict[str, str | int],
    **kwargs
):
    dynamo.get_table(table).delete_item(Key=kv, **kwargs)


def _scan(
    dynamo: DynamoAccess,
    table_name: str,
    **kwargs
) -> list[Any]:
    items = []
    last_evaluated_key = None
    table = dynamo.get_table(table_name)

    while True:
        if last_evaluated_key:
            response = table.scan(ExclusiveStartKey=last_evaluated_key, **kwargs)
        else:
            response = table.scan(**kwargs)

        items.extend(response['Items'])
        last_evaluated_key = response.get('LastEvaluatedKey')

        if not last_evaluated_key:
            break

    return prepare_item(items)  # type: ignore


class Model:
    @staticmethod
    def make_id() -> str:
        return str(uuid.uuid4())


def without(d, keys: list[str]):
    return {k: v for k, v in d.items() if k not in keys}


class DbException(Exception):
    pass


class DbNotFound(DbException):
    pass



import os
import boto3

from prismarine.runtime.dynamo_access import DynamoAccess


class MyDynamoAccess(DynamoAccess):
    def get_resource(self):
        return boto3.resource('dynamodb')

    def get_table(self, full_model_name: str):
        env = os.getenv('ENV')
        return self.get_resource().Table(f'{full_model_name}-{env}')


dynamoaccess = MyDynamoAccess()


def get_dynamo_access():
    return dynamoaccess



from pathlib import Path
import logging as lg
from importlib.metadata import version

import click

from prismarine.prisma_common import set_path
from prismarine.prisma_client import generate_client


@click.group()
@click.pass_context
@click.option(
    '--path', multiple=True, type=click.Path(exists=True),
    help='Additional Python path to use'
)
@click.option(
    '--verbose', is_flag=True, help='Enable verbose logging'
)
def prismarine(ctx, path, verbose):
    ctx.ensure_object(dict)
    lg.basicConfig(level=lg.DEBUG if verbose else lg.INFO)
    paths = list(Path(p).resolve() for p in path) if path else []
    set_path(paths)


@prismarine.command(
    name='generate-client',
    help='Generate a Prismarine client сщву for a given cluster package'
)
@click.option(
    '--base', required=True, type=click.Path(exists=True),
    help="Primary Python path to use while searching for 'models' package"
)
@click.option(
    '--runtime', required=False,
    help='Parent package for models to use in runtime. If not provided, assume top-level'
)
@click.option(
    '--dynamo-access-module', required=False,
    help='''
Dynamo access module to use in runtime. If not provided, DefaultDynamoAccess access class from prismarine.runtime.dynamo_default will be used'''
)
@click.argument('cluster_package')
def generate_client_cmd(base, runtime, dynamo_access_module, cluster_package):
    base_dir = Path(base)

    generate_client(
        base_dir,
        cluster_package,
        runtime,
        dynamo_access_module
    )


@prismarine.command(
    name='version',
    help='Print the version of Prismarine'
)
def version_cmd():
    click.echo(version('prismarine'))


def main():
    prismarine()


if __name__ == '__main__':
    main()



from typing import TypedDict, NotRequired
from prismarine.runtime import Cluster


c = Cluster('MyDatabase')


@c.model(PK='Foo', SK='Bar')
class Item(TypedDict):
    Foo: str
    Bar: str
    Baz: NotRequired[str]



def build_dynamo_tables(prefix, cluster):
    'Returns: (Full Table Name with Prefix, EasySAM Table Definition)'

    result = {}

    for _, data in cluster.models.items():
        table_name = data['table']
        short_name = table_name.replace(prefix, '')

        main = data['main']

        attributes = {}
        indices = []

        if 'PK' not in main or not main['PK']:
            raise UserWarning(f'PK is required for table {table_name}')

        attributes[main['PK']] = {'hash': True}

        if 'SK' in main and main['SK']:
            attributes[main['SK']] = {'range': True}

        for index_name, index in data['indexes'].items():
            sam_index = {'name': index_name}
            index_attributes = []

            if 'PK' not in index or not index['PK']:
                raise UserWarning(f'PK is required for index {index} in table {table_name}')

            index_attributes.append({'name': index['PK'], 'hash': True})

            if index['PK'] not in attributes:
                attributes[index['PK']] = {}

            if 'SK' in index and index['SK']:
                if index['SK'] not in attributes:
                    attributes[index['SK']] = {}

                index_attributes.append({'name': index['SK'], 'range': True})

            sam_index['attributes'] = index_attributes
            indices.append(sam_index)

        def named_dict_to_item(name, data):
            result = {'name': name}
            result.update(data)
            return result

        result[short_name] = {
            'attributes': list(named_dict_to_item(k, v) for k, v in attributes.items()),
        }

        if indices:
            result[short_name]['indices'] = indices

    return result



# This file is generated by prismarine. Do not modify it.

from prismarine.runtime.dynamo_crud import (
    _query,
    _get_item,
    _update,
    _put_item,
    _delete,
    _scan,
    _save,
    Model,
    without,
    DbNotFound
)

from typing import TypedDict, List, Literal, NotRequired  # noqa
from types import EllipsisType

from prismarine.runtime.dynamo_default import get_dynamo_access
from myobject.models import Item

dynamo = get_dynamo_access()


class ItemModel(Model):
    table_name = 'MyDatabaseItem'
    PK = 'Foo'
    SK = 'Bar'

    class UpdateDTO(TypedDict, total=False):
        Foo: str
        Bar: str
        Baz: NotRequired[str]

    @staticmethod
    def list(*, foo: str, **kwargs) -> List[Item]:
        return _query(dynamo, ItemModel.table_name, {ItemModel.PK: foo}, **kwargs)

    @staticmethod
    def get(
        *, bar: str, foo: str, default: Item | EllipsisType = ..., **kwargs
    ) -> Item:
        item = _get_item(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            default,
            **kwargs
        )
        if not item:
            raise DbNotFound(
                f'Item was not found in {ItemModel.table_name} with params foo={foo}, bar={bar}'
            )
        return item

    @staticmethod
    def put(item: Item, **kwargs) -> Item:
        _put_item(dynamo, ItemModel.table_name, item, **kwargs)
        return item

    @staticmethod
    def update(
        item: UpdateDTO, *, foo: str, bar: str, default: Item | EllipsisType = ...
    ) -> Item:
        initial = ItemModel.get(bar=bar, foo=foo, default=default)
        _update(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            without(item, [ItemModel.PK, ItemModel.SK])
        )
        return {**initial, **item}

    @staticmethod
    def save(updated: Item, *, original: Item | None = None) -> Item:
        _save(
            dynamo,
            ItemModel.table_name,
            kv={
                ItemModel.PK: updated[ItemModel.PK],
                ItemModel.SK: updated[ItemModel.SK]
            },
            updated=updated,
            original=original
        )
        return updated

    @staticmethod
    def delete(*, bar: str, foo: str, **kwargs):
        _delete(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            **kwargs
        )

    @staticmethod
    def scan(**kwargs) -> List[Item]:
        return _scan(dynamo, ItemModel.table_name, **kwargs)
